"""
Specific types or type aliases used in the library.
"""

from __future__ import annotations
import os
import typing as tp
import pathlib
import argparse

import matplotlib
import pandas
import numpy
from anndata import AnnData as _AnnData


__all__ = [
    "Path",
    "GenericType",
    "Args",
    "Array",
    "MultiIndexSeries",
    "Series",
    "DataFrame",
    "AnnData",
    "Figure",
    "Axis",
    "Patch",
    "ColorMap",
]


class Path(pathlib.Path):
    """
    A pathlib.Path child class that allows concatenation with strings
    by overloading the addition operator.

    In addition, it implements the ``startswith`` and ``endswith`` methods
    just like in the base :obj:`str` type.

    The ``replace_`` implementation is meant to be an implementation closer
    to the :obj:`str` type.

    Iterating over a directory with ``iterdir`` that does not exists
    will return an empty iterator instead of throwing an error.

    Creating a directory with ``mkdir`` allows existing directory and
    creates parents by default.
    """

    _flavour = (
        pathlib._windows_flavour  # type: ignore[attr-defined]  # pylint: disable=W0212
        if os.name == "nt"
        else pathlib._posix_flavour  # type: ignore[attr-defined]  # pylint: disable=W0212
    )

    def __add__(self, string: str) -> Path:
        return Path(str(self) + string)

    def startswith(self, string: str) -> bool:
        return str(self).startswith(string)

    def endswith(self, string: str) -> bool:
        return str(self).endswith(string)

    def replace_(self, patt: str, repl: str) -> Path:
        return Path(str(self).replace(patt, repl))

    def iterdir(self) -> tp.Generator:
        if self.exists():
            yield from [Path(x) for x in pathlib.Path(str(self)).iterdir()]
        yield from []

    def unlink(self, missing_ok: bool = True) -> Path:
        super().unlink(missing_ok=missing_ok)
        return self

    def mkdir(self, mode=0o777, parents: bool = True, exist_ok: bool = True) -> Path:
        super().mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        return self

    def glob(self, pattern: str) -> tp.Generator:
        # to support ** with symlinks: https://bugs.python.org/issue33428
        from glob import glob

        if "**" in pattern:
            sep = "/" if self.is_dir() else ""
            yield from map(
                Path,
                glob(self.as_posix() + sep + pattern, recursive=True),
            )
        else:
            yield from super().glob(pattern)


GenericType = tp.TypeVar("GenericType")

# type aliasing (done with Union to distinguish from other declared variables)


# # Args = Union[argparse.Namespace]
# class Args(argparse.Namespace, tp.Mapping[str, tp.Any]):
#     pass


# # Series = Union[pandas.Series]
# class Series(pandas.Series, tp.Mapping[tp.Any, tp.Any]):
#     pass


Args = tp.Union[argparse.Namespace, tp.Mapping[str, tp.Any]]

Array = numpy.ndarray

MultiIndexSeries = pandas.Series
Series = pandas.Series
DataFrame = pandas.DataFrame
AnnData = _AnnData

# Figure = matplotlib.figure.Figure
# Axis = matplotlib.axis.Axis
# Patch = matplotlib.patches.Patch
# ColorMap = matplotlib.colors.LinearSegmentedColormap


# Tests
from imc_analysis.tl import impute_roi_area
import imc_analysis as imc
from imc_analysis.logging import *
from tqdm import tqdm
logger.setLevel(logging.DEBUG)


if __name__ == '__main__':
    
    logger.info('Downloading sample file')
    
    import scanpy as sc
    adata = sc.read(
        'data/healthy_lung_adata.h5ad',
        backup_url='https://zenodo.org/record/6376767/files/healthy_lung_adata.h5ad?download=1')

    logger.debug(adata)



    for ct in ['cell type']:
        for ck in [[], ['Age']]:
            logger.debug(f"Condition: {ck}")
            # roi density
            density = imc.tl.celltype_density(adata, celltype = ct, condition_keys = ck)
            # patient density
            p_density = imc.tl.patient_density(adata, patient_key = 'Patient', celltype_key = ct, condition_keys = ck)
            # roi mean
            logger.debug(f"Measuring ROI Mean")
            roi_mean = imc.tl.compute_mean(adata, obs_key = 'roi', condition_keys = ck)
            # patient mean
            logger.debug(f"Measuring Patient Mean")
            p_mean = imc.tl.compute_mean(adata, obs_key = 'Patient', condition_keys = ck)
            
            for data in tqdm([density, p_density, roi_mean, p_mean]):
                imc.tl.grouped_mwu_test(data, condition_keys = ck)


from . import pl
from . import tl
from . import utils

from .pl import celltype_heatmap, plot_mwu, umap_var
from .tl import celltype_density, patient_density, compute_mean, grouped_mwu_test
from .utils import parse_yaml

import scanpy as sc
import matplotlib as mpl

sc.settings.set_figure_params(dpi=200, dpi_save=300, fontsize=12)
mpl.rcParams["pdf.fonttype"] = 42
mpl.rcParams["ps.fonttype"] = 42
mpl.rcParams["axes.grid"] = False


from imc_analysis.types import Path
from imc_analysis.logging import *
import typing as tp

def parse_yaml(yaml_file: Path) -> tp.Union[dict, None]:
    """
    Parse project config yaml file
    """
    import yaml
    logger.info(f'Opening project config file: {yaml_file}...')
    with open(yaml_file, "r") as stream:
        try:
            metadata = yaml.safe_load(stream)
            logger.info(f'Successfully opened project config file: {yaml_file}...')
            return metadata
        except yaml.YAMLError as exc:
            logger.error(f'Error opening config file: {yaml_file}')
            print(exc)

def download(url: str, save: Path) -> tp.Union[dict, None]:
    """
    Download file from url
    """
    # if metadata not found, download
    import os

    file_path = Path(save)  # Replace with your filename
    directory = file_path.resolve().parent  # Get absolute directory

    if not os.path.exists(file_path):
        from tqdm import tqdm
        import requests

        os.makedirs(directory, exist_ok = True)
        response = requests.get(url, stream=True)
        
        with open(save, "wb") as handle:
            for data in tqdm(response.iter_content()):
                handle.write(data)
            handle.close()


def init(project, yaml, panel):
    """
    Create a IMC project directory
    """
    import os

    os.makedirs(f'{project}/data', exist_ok = True)
    os.makedirs(f'{project}/processed', exist_ok = True)
    os.makedirs(f'{project}/metadata', exist_ok = True)
    os.makedirs(f'{project}/figures', exist_ok = True)
    os.makedirs(f'{project}/images', exist_ok = True)

    yaml_str = f"---\n# {project} Project Config YAML File\n\n"

    for p in panel:
        yaml_str += f"""\
{p}:
    AnnData:
        unlabeled_file_name: 'results/unlabeled.h5ad'
        labeled_file_name: 'results/clustered.h5ad'
        phenotyped_file_name: 'results/celltyped.h5ad'
        phenotyped_umap_name: 'results/celltyped.umap.h5ad'
        utag_file_name: 'results/celltyped.utag.h5ad'

    glob_pattern:
        IMAGE_FILE_PATTERN: 'processed/{p}/*/tiffs/*_full.tiff'
        MASK_FILE_PATTERN: 'processed/{p}/*/tiffs/*_full_mask.tiff'
        NUCMASK_FILE_PATTERN: 'processed/{p}/*/tiffs/*_full_nucmask.tiff'
        CSV_FILE_PATTERN: 'processed/{p}/*/tiffs/*_full.csv'

    image_metadata:
        ROI_AREA_FILE: 'metadata/{p}_ROI_area.csv'
        ROI_MEAN_FILE: 'metadata/{p}_ROI_mean.csv'
        ROI_VAR_FILE: 'metadata/{p}_ROI_var.csv'
        ROI_NUC_FILE: 'metadata/{p}_ROI_nuc.csv'
        ROI_CYTO_FILE: 'metadata/{p}_ROI_cyto.csv'
        ROI_SPILLOVER_FILE: 'metadata/{p}_ROI_spillover.csv'

    UMAP_FRACTION: 0.1

    var_celltype_groups:
        'Epithelial': ['PanCytokeratin(Pt198)']
        'Stromal': ['CD31(Eu151)', 'aSMA(Pt196)', 'Vimentin(Nd143)']
        'Prolif.': ['Ki67(Er168)']

        'Mono.': ['HLADR(Pr141)',  'CD14(Nd144)',]
        'Mac.': ['CD14(Nd144)', 'CD16(Nd146)', 'CD68(Tb159)']

        'T': ['CD3(Er170)', 'CD45(Sm152)', 'CD27(Yb171)',]
        'CD4': [ 'CD4(Gd156)']
        'CD8': ['CD8a(Dy162)',]
        'B': ['CD20(Dy161)',]
        'NK': ['CD11b(Dy164)', 'CD56(Dy163)', 'CD57(Tm169)',]

    plot_config : {{
        'Epithelial': {{
            38: ['KRT5', 0.1],
            31: ['KRT818', 0.01],
            10: ['ECad', 0.1],
            46: ['DNA', 0.003]
        }}
    }},

"""
    with open(f'{project}/{yaml}', 'w') as file:
        file.write(yaml_str)
        file.close()




import logging

# logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
# logger = logging.getLogger(__name__)
# #logger.setLevel(logging.INFO)  # Set the log level
# logger.setLevel(logging.DEBUG)  # Set the log level

# # set logging level for fonttools to warning
# logging.getLogger("fontTools").setLevel(logging.WARNING)

# # Create a console handler and set its level and format
# console_handler = logging.StreamHandler()
# #console_handler.setLevel(logging.INFO)
# console_handler.setLevel(logging.DEBUG)
# formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

# console_handler.setFormatter(formatter)

# # Add the console handler to the logger
# logger.addHandler(console_handler)


# Create a custom logger for this module
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)  # Set logger's level to capture DEBUG and above

# Remove any existing handlers (to avoid duplicate logging)
if logger.hasHandlers():
    logger.handlers.clear()

# Create a console handler with the same log level
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)

# Create a formatter and set it for the handler
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
console_handler.setFormatter(formatter)

# Add the handler to the custom logger
logger.addHandler(console_handler)
logger.propagate = False

# Configure external libraries as needed
logging.getLogger("fontTools").setLevel(logging.WARNING)


# Tests
import imc_analysis as imc
from imc_analysis.logging import *

from tqdm import tqdm
logger.setLevel(logging.DEBUG)


if __name__ == '__main__':
    
    logger.info('Loading sample file')
    
    import scanpy as sc
    adata = sc.read(
        'data/infected_lung_adata.h5ad',
        backup_url = 'https://zenodo.org/records/6376767/files/infected_lung_adata.h5ad?download=1')

    import numpy as np
    adata.obsm['spatial'] = np.array(adata.obs[['Y', 'X']])
    logger.debug(adata)
    # series of tests to ensure build works

    imc.pl.celltype_heatmap(adata, cluster_ids = ['cluster_1.0', 'cluster_label', 'metacluster_label'], out_dir='figures/celltype/heatmap/')
    imc.pl.umap_var(adata, outdir = 'figures/umap/')

    for ct in ['cluster_1.0', 'cluster_label', 'metacluster_label']:
        conditions = ['disease', 'phenotypes']
        
        # plot roi
        for roi in adata.obs['roi'].unique()[:3]:
            a = adata[adata.obs['roi'] == roi]
            imc.pl.spatial(a, color = ct, edges = None)
            imc.pl.spatial(a, color = ct, edges = 'KNN', edges_knn = 15)
            imc.pl.spatial(a, color = ct, edges = 'radius', edges_radius = 40)


        # roi density
        density = imc.tl.celltype_density(adata, celltype = ct, condition_keys = conditions, x_coord = 'X', y_coord = 'Y')

        # roi mean
        roi_mean = imc.tl.compute_mean(adata, obs_key = 'roi', condition_keys = conditions)
        
        for data in tqdm([density, roi_mean, ]):
            imc.tl.grouped_mwu_test(data, condition_keys = conditions)
            
            logger.info('Plotting MWU plots')            
            for pval_form in ['star', 'sci_not']:
                imc.pl.plot_mwu(data, save_dir = f'figures/{ct}/', palette = None, pval_form = pval_form)
            
            


# main.py

import click
from imc_analysis.utils import parse_yaml
import os

@click.group()
def cli():
    """A Imaging Mass Cytometry analysis command line tool. Current commands include qc, phenotype, and visualize."""

@cli.command()
@click.option('-p', '--project', 'project', type=click.Path(), default = 'imc_project/')
@click.option('-y', '--yaml', 'yaml', type=click.Path(), default = 'metadata/project.yml', )
@click.option('--panel','panel', type=list, default = ['PANEL'])
def init(project, yaml, panel):
    """
    Create a IMC project directory
    """
    from imc_analysis.utils import init

    init(project, yaml, panel)


@cli.command()
@click.option('-y', '--yaml', 'yaml', type=click.Path(exists=True), default = 'metadata/project.yml')
@click.option('-p', '--panel','panel', type=str)
@click.option('-o', '--outdir','outdir', type=click.Path(exists=True), default='figures/QC/')
def qc(yaml, panel, outdir):
    """
    Run quality control check on an IMC project.

    yaml: Path to yaml file storing project configs.
    panel: Panel to apply qc 
    """
    click.echo(f"Plotting QC figures using metadata stored in {yaml}.")
    metadata = parse_yaml(yaml)
    assert(panel in metadata)

    from imc_analysis.qc.qc import quality_control

    path = f'{outdir}/{panel}'
    os.makedirs(path, exist_ok=True)
    quality_control(metadata[panel], outdir = path)

    click.echo(f"Command QC completed")

@cli.command()
@click.option('-y', '--yaml','yaml', type=click.Path(exists=True), default = 'metadata/project.yml')
@click.option('-p', '--panel', 'panel', type=str)
def phenotype(yaml):
    """Analyze phenotypes."""
    from imc_analysis.phenotype.phenotype import phenotype
    phenotype()

@cli.command()
@click.option('-y', '--yaml','yaml', type=click.Path(exists=True), default = 'metadata/project.yml')
@click.option('-m', '--mode','mode', type=click.Choice(['stack', 'rgb']), default = 'stack')
@click.option('-p', '--panel', 'panel', type=str)
@click.option('-o', '--outdir','outdir', type=click.Path(exists=True), default='images/')
def visualize(mode, yaml, panel, outdir):
    """Visualize data."""

    metadata = parse_yaml(yaml)

    from imc_analysis.visualize.visualize import visualize_stack
    
    path = f'{outdir}/{mode}/{panel}/'
    os.makedirs(path, exist_ok=True)
    
    visualize_stack(metadata, panel, outdir = path, mode = mode)


@cli.command()
@click.option('-y', '--yaml', 'yaml', type=click.Path(exists=True), default = 'metadata/project.yml')
@click.option('-p', '--panel','panel', type=str)
def celltype(yaml):
    """Perform cell phenotyping"""

@cli.command()
@click.option('-y', '--yaml', 'yaml', type=click.Path(exists=True), default = 'metadata/project.yml')
@click.option('-p', '--panel','panel', type=str)
def stats(yaml):
    """Perform statistical testing either on single cell level, ROI level, or Patient level"""


@cli.command()
@click.option('-y', '--yaml', 'yaml', type=click.Path(exists=True), default = 'metadata/project.yml')
@click.option('-p', '--panel', 'panel', type=str)
@click.option('-m', '--mode', 'mode', type=click.Choice(['roi', 'patient']), default = 'roi')
@click.option('-k', '--key', 'key', type=str, default = 'roi')
def agg(yaml, panel, mode, key):
    """Aggregate single cell matrix on ROI level or Patient level"""
    click.echo(f"Aggregate {panel} single cell matrix on {mode} level {yaml}.")
    metadata = parse_yaml(yaml)
    assert(panel in metadata)

    
if __name__ == "__main__":
    cli()



from setuptools import setup
setup()

# from setuptools import setup

# setup(
#     name="imc-analysis",
#     version="0.1.7",
#     description="A multi-file Python command line tool with commands qc, phenotype, and visualize.",
#     author="Junbum Kim",
#     author_email="jkim0810@gmail.com",
#     url="https://github.com/jkim810/imc-analysis",
#     packages=["imc_analysis"],
#     package_dir={"": "."},
#     include_package_data=True,
#     install_requires=[
#         "click",
#         "pathlib",
#         "tifffile==2023.4.12",
#         "tqdm",
#         "pandas==2.0.1",
#         "scanpy==1.9.3",
#         "anndata==0.9.1",
#         "scikit-image==0.20.0",
#         "scipy==1.9.1",
#         "seaborn",
#         "pyyaml==6.0",
#         "pydot==1.4.2",
#     ],
#     entry_points = {
#         'console_scripts': ['mybinary=imc_analysis.main:cli'],
#     },
#     test_suite="tests",
#     python_requires=">=3.6",
# )



from tqdm import tqdm
from glob import glob
import tifffile
import matplotlib.pyplot as plt

import typing as tp
from imc_analysis.types import Path, AnnData

import os

import pandas as pd
import numpy as np

import scanpy as sc
import anndata

from skimage.exposure import adjust_gamma
from skimage import filters
import scipy.ndimage as ndi
import scipy

import seaborn as sns
import matplotlib.pyplot as plt

import warnings
warnings.simplefilter("ignore", UserWarning)
from imc_analysis.logging import *

# import matplotlib

# # matplotlib.use('Agg')
# sc.settings.set_figure_params(dpi=200, dpi_save=300, fontsize=12)
# matplotlib.rcParams["pdf.fonttype"] = 42
# matplotlib.rcParams["ps.fonttype"] = 42
# matplotlib.rcParams["axes.grid"] = False

def plot_mwu(
    adata: AnnData,
    line_width:float = 0.5,
    save_dir: Path = 'figures/',
    kind = 'box', # either box or bar
    pval_form: str = 'star',
    show: bool = False,
    nrow = 3,
    ncol = 6,
    y_max = None,
    figsize = (12,8),
):

    if 'mwu' not in adata.uns:
        logger.error("'mwu' not found in adata.uns layer. Run imc_analysis.tl.grouped_mwu_test()")
        return
    
    density = adata.to_df()
    mwu_ = adata.uns['mwu']
    # for each conditions
    for cond in tqdm(mwu_['condition'].unique()):
        
        logger.info(f'Producing figure for condition: {cond}')
        
        fig, axes = plt.subplots(nrow,ncol, dpi = 300, figsize = figsize)
        
        if f'{cond}_colors' in adata.uns:
            palette = adata.uns[f'{cond}_colors']
        else:
            palette = None #'tab10'

        # for each celltype
        for i, ax in enumerate(axes.flatten()):
        
            if i >= density.shape[1]:
                ax.axis('off')
            else:
                
                ct = density.columns[i]
                # create swarmboxen plot
                if kind == 'box':
                    sns.boxplot(y = density[ct], x = adata.obs[cond], hue = adata.obs[cond], ax = ax, fliersize = 0, palette = palette, dodge = False)
                    sns.swarmplot(y = density[ct], x = adata.obs[cond], color = 'black', ax = ax, s = 3)
                elif kind == 'box-line':
                    sns.boxplot(y = density[ct], x = adata.obs[cond], hue = adata.obs[cond], ax = ax, fliersize = 0, palette = palette, dodge = False, boxprops=dict(alpha=.65,), whiskerprops=dict(alpha=0.5,), capprops = dict(alpha = 0.7), medianprops = dict(alpha=0.7,), meanprops = dict(alpha=0.7,), width = 0.7)
                    sns.swarmplot(y = density[ct], x = adata.obs[cond], color = 'black', ax = ax, s = 2, alpha=0.5)
                    sns.lineplot(y = density[ct], x = adata.obs[cond], ax = ax, color = 'black', estimator='median', markers=True, dashes=False, linewidth = 2, marker='o', markeredgecolor='black')
                elif kind == 'bar':
                    sns.barplot(y = density[ct], x = adata.obs[cond], hue = adata.obs[cond], ax = ax, palette = palette, dodge = False, width = 0.7)
                elif kind == 'violin':
                    sns.violinplot(y = density[ct], x = adata.obs[cond], hue = adata.obs[cond], ax = ax, palette = palette, dodge = False, width = 0.7)
                else:
                    logger.error("'kind' must equal either 'box', 'bar', or 'violin'")
                    return
                ax.set_title(ct)
                ax.set_ylabel('')
                ax.set_xlabel('')

                y1 = density[ct].max() * 1.05
                if y_max:
                    ax.set_ylim(0, y_max)
                    y1 = y_max * 0.85
                r = y1 * 0.03
                l = adata.obs[cond].cat.categories.tolist()

                sig_mwu_ = mwu_[(mwu_['celltype'] == ct) & (mwu_['condition'] == cond) & (mwu_['adj. p-val'] < 0.05)]

                sig_n = 0
                for i, idx in enumerate(sig_mwu_.index):
                    pair = sig_mwu_.loc[idx, 'pair']
                    
                    if pval_form == 'star':
                        pval = sig_mwu_.loc[idx, 'significance']
                    else:
                        pval = sig_mwu_.loc[idx, 'adj. p-val sci']
                    
                    if len(sig_mwu_) == 1:

                        ax.plot([pair[0], pair[1]], [y1 + r*i, y1 + r*i], lw=line_width, c='black')
                        ax.text(s = pval, x = 0.5, y = y1, fontsize = 8, va = 'bottom', ha = 'center')

                    else:
                        
                        ax.plot([pair[0], pair[1]], [y1 + r*sig_n, y1 + r*sig_n], lw=line_width, c='black')
                        ax.text(s = pval, x = pair[1], y = y1 + r*sig_n, fontsize = 8, va = 'top', ha = 'left')
                        sig_n += 1
                ax.legend().remove()
                sns.despine()
        plt.tight_layout()

        dir_path = f'{save_dir}/{cond}_{pval_form}.pdf'
        # check if directory exists
        if not os.path.exists(save_dir):
            # create directory if it doesn't exist
            os.makedirs(save_dir)
            print(f"Directory '{save_dir}' created.")
        plt.savefig(dir_path, bbox_inches = 'tight')
        
        if show:
            plt.show()
        
        plt.close()

def celltype_heatmap(
    adata: anndata.AnnData,
    var_names: dict = None,
    umap_adata: anndata.AnnData = None,
    cluster_ids: list = ['cluster_0.5', 'cluster_1.0', 'cluster_1.5', 'cluster_2.5', 'celltype', 'celltype_cid', 'celltype_broad'],
    panel: str = 'panel',
    plot_umap: bool = True,
    cmap: str = 'Spectral_r',
    out_dir: Path = 'figures/cell phenotyping/'
):
    
    os.makedirs(out_dir, exist_ok = True)
    if var_names == None:
        logger.info(f"'var_names' was not provided. Defaulting to original var indices")
        var_names = adata.var.index
        
    for cluster_id in cluster_ids:
        
        if cluster_id not in adata.obs:
            logger.info(f'{cluster_id} not in "adata.obs" layer. Skipping...')
            continue

        # Plot Raw Matrixplot
        mp = sc.pl.matrixplot(
            adata,
            groupby = cluster_id,
            var_names = var_names,
            log = True,
            cmap = cmap,
            vmax = 1.5,
            return_fig = True,
            use_raw = False
        )
        mp.add_totals().style(cmap = cmap, edge_color='black')
        mp.savefig(f'{out_dir}/matrixplot_raw_{cluster_id}.pdf')
        plt.close()

        # Plot Normalized Matrixplot
        mp = sc.pl.matrixplot(
            adata,
            groupby = cluster_id,
            var_names = var_names,
            log = True,
            cmap = cmap,
            standard_scale = 'var',
            return_fig = True,
            use_raw = False
        )
        mp.add_totals().style(cmap = cmap, edge_color='black')
        mp.savefig(f'{out_dir}/matrixplot_normalized_{cluster_id}.pdf')
        plt.close()

        if plot_umap and 'X_umap' in adata.obsm:
            fig = sc.pl.umap(adata, color = cluster_id, frameon = False, title = '', show = False, return_fig = True, use_raw = False)
            plt.tight_layout()
            fig.savefig(f'{out_dir}/umap_{cluster_id}.pdf', bbox_inches = 'tight')
            plt.close()
        
    if 'celltype' in adata.obs:
        adata.obs.groupby(['celltype']).count().plot.pie(y = 'sample', autopct = '%1.1f%%', cmap = 'PiYG')
        plt.legend().remove()
        plt.ylabel('')
        plt.savefig(f'{out_dir}/celltype_proportion.pdf', bbox_inches = 'tight')
        plt.close()

def spatial(
    adata: anndata.AnnData,
    color: tp.Union[str, list],
    spot_size: int = 10,
    edges: tp.Any = None,
    edges_knn: int = 15,
    edges_radius: float = 40,
    *args,
    **kwargs,
):
    ''' Function to plot spatial plots from AnnData. Spatial coordinates must be stored in adata.uns['spatial'] for this function to work.

    Args:
        adata (anndata.AnnData): AnnData object to spatial plot
        color (tp.Union[str, list]): color of points in spatial plot. Variable must be stored in 'obs' layer
        spot_size (int): default size of each cells
        edges (tp.Any): Whether to connect cells with edges. Default is to not plot edges. Two options are 'KNN' and 'radius'. If 'KNN', numbers of neighbors are defined by 'edges_knn'. If 'radius', number of neighbors are defined by 'edges_radius'.
        edges_knn (int): number of nearest neighbors in KNN edges
        edges_radius (float): threshold distance for neighborhood definition    '''
    import squidpy as sq
    
    a = adata.copy()

    if edges == 'KNN':
        sq.gr.spatial_neighbors(a, coord_type = 'generic', n_neighs = edges_knn)
    elif edges == 'radius':
        sq.gr.spatial_neighbors(a, coord_type = 'generic', radius = edges_radius)
    else:
        logger.info("'edges' is neither 'KNN', nor 'radius'. Plotting figure without edges")
        sc.pl.spatial(
            a,
            color = [color],
            spot_size = spot_size,
            frameon = False,
            show = False,
            *args,
            **kwargs,
        )

        return

    sc.pl.spatial(
        a,
        color = [color],
        spot_size = spot_size,
        edges = True,
        neighbors_key = 'spatial_neighbors',
        frameon = False,
        show = False,
        *args,
        **kwargs
    )

def add_scale_box_to_fig(
    img,
    ax,
    box_width: int = 100,
    box_height: float = 3,
    color: str = 'white'
):    
    import matplotlib.patches as patches
    x = img.shape[1]
    y = img.shape[0]
    
    # Create a Rectangle patch
    rect = patches.Rectangle((x - box_width, y * (1-box_height/100)), box_width, y * (box_height/100), linewidth=0.1, edgecolor='black', facecolor=color)
    
    
    # Add the patch to the Axes
    ax.add_patch(rect)
    return ax

def visualize_grayscale(
    img_name: str,
    csv_name: str,
    out_dir: str = 'images'
) -> None:
    fig_dict = {
        'nrow': [5, 6, 6, 8],
        'ncol': [8, 8, 10, 10],
        'figsize': [(20,15), (20,15), (20,15), (25,20)],
    }
    
    img = tifffile.imread(img_name)
    roi = img_name.split('/')[-1].replace('_full.tiff', '')
    df = pd.read_csv(csv_name, index_col = 0)
    df['channel'] = df['channel'].str.replace('[0-9]{2,3}[A-Z][a-z]', '', regex = True)
    n_feature = df.shape[0]

    for i in range(4):
        if n_feature < fig_dict['nrow'][i] * fig_dict['ncol'][i]:
            break

    fig, axs = plt.subplots(
        fig_dict['nrow'][i],
        fig_dict['ncol'][i],
        figsize = fig_dict['figsize'][i],
        dpi = 300
    )
    
    for i, ax in enumerate(fig.axes):

        if i < len(df):
            rescaled_img = adjust_gamma(img[i], gamma = 0.2, gain = 1)

            ax.imshow(rescaled_img, cmap = 'viridis')

            ax.set_title(df.iloc[i]['channel'], fontsize = 10)
            add_scale_box_to_fig(rescaled_img, ax, box_width = 200)

            ax.annotate(
                text = 'min: {:.2f}\nmax: {:.2f}'.format(img[i].min(), img[i].max()),
                xy = (rescaled_img.shape[1], 0),
                ha = 'right',
                va = 'bottom',
                fontsize = 6
            )
            
            ax.annotate(
                text = '200μm',
                xy = (rescaled_img.shape[1], rescaled_img.shape[0]),
                ha = 'right',
                va = 'top',
                fontsize = 6
            )
        ax.axis('off')

    plt.suptitle(roi)
    plt.tight_layout()
    plt.savefig(f'{out_dir}/{roi}.pdf', bbox_inches = 'tight')
    plt.savefig(f'{out_dir}/{roi}.png', bbox_inches = 'tight')
    plt.close()

def visualize_rgb(
    metadata: dict,
    image_name: Path,
    csv_name: Path,
    outdir: Path,
    plot_config_string: str = 'plot_config',
):
    
    assert(plot_config_string in metadata)
    plot_config = metadata[plot_config_string]
    plot_keys = list(plot_config.keys())

    try:
        image = tifffile.imread(img_name)
        csv = pd.read_csv(csv_name)
    except FileNotFoundError:
        print(f'Either {img_name} or {csv_name} not found')

    img_name = img_name.split('/')[-1]
    fig, axes = plt.subplots(2,4,dpi=300, figsize = (24,12))

    for i, ax in enumerate(axes.flatten()):
        
        if i < len(plot_keys):
            plot_key = plot_keys[i]
            image_sub = [image[k] * float(v[1]) for k, v in plot_config[plot_key].items()]
            image1 = np.clip(np.stack(image_sub[:-1], axis = 2) + image_sub[-1][:,:,np.newaxis], 0, 1)
            image1[-int(image.shape[1] * 0.03):, -200:, :] = 1
            title = plot_key + '\n'
            for idx, key in enumerate(plot_config[plot_key]):
                title += f'{plot_config[plot_key][key][0]}, '
            ax.imshow(image1)
            ax.set_title(title[:-2])
            ax.axis('off')

    filename = img_name.replace('.tiff','')
    title = filename

    plt.suptitle(title, fontsize = 14)
    plt.tight_layout()
    plt.savefig(Path(f'{out_dir}/{filename}_RGB.pdf'), bbox_inches = 'tight')
    plt.savefig(Path(f'{out_dir}/{filename}_RGB.png'), bbox_inches = 'tight')
    
    plt.close()

def umap_var(
    adata: anndata.AnnData = None,
    metadata: dict = None,
    anndata_key: str = None,
    outdir: Path = 'figures/',
):
    if metadata != None and anndata_key != None:
        adata = sc.read(metadata[anndata_key])

    assert(isinstance(adata, anndata.AnnData))
    os.makedirs(outdir, exist_ok=True)
    # Plot UMAP Marker
    print('Plotting marker UMAP')
    fig_dict = {
        'nrow': [5, 6, 6, 8],
        'ncol': [8, 8, 10, 10],
        'figsize': [(20,15), (20,15), (20,15), (25,20)],
    }
    
    for i in range(4):
        if len(adata.var) < fig_dict['nrow'][i] * fig_dict['ncol'][i]:
            break

    fig, axs = plt.subplots(
        fig_dict['nrow'][i],
        fig_dict['ncol'][i],
        figsize = fig_dict['figsize'][i],
        dpi = 300
    )

    for i, ax in tqdm(enumerate(fig.axes)):
        if i < len(adata.var.index):
            var = adata.var.index[i]
            sc.pl.umap(
                adata,
                color = var,
                use_raw = False,
                size = 1,
                frameon = False,
                ax = ax,
                show = False,
                colorbar_loc = None,
                vmin = 0,
                vmax = 3
            )
            
        else:
            ax.axis('off')
            
    plt.suptitle(f'Cell Marker Scaled Expression UMAP\nPlotting {len(adata)} cells')
    plt.tight_layout()
    
    plt.savefig(
        f'{outdir}/umap_marker_normalized.pdf',
        bbox_inches = 'tight'
    )
    plt.close()


    for i in range(4):
        if len(adata.var) < fig_dict['nrow'][i] * fig_dict['ncol'][i]:
            break

    fig, axs = plt.subplots(
        fig_dict['nrow'][i],
        fig_dict['ncol'][i],
        figsize = fig_dict['figsize'][i],
        dpi = 300
    )

    for i, ax in tqdm(enumerate(fig.axes)):
        if i < len(adata.var.index):
            var = adata.var.index[i]
            sc.pl.umap(
                adata,
                color = var,
                use_raw = True,
                size = 1,
                frameon = False,
                ax = ax,
                show = False,
                colorbar_loc = None,
            )
            
        else:
            ax.axis('off')
            
    plt.suptitle(f'Cell Marker Raw Expression UMAP\nPlotting {len(adata)} cells')
    plt.tight_layout()
    plt.savefig(
        f'{outdir}/umap_marker_raw.pdf',
        bbox_inches = 'tight'
    )
    plt.close()



from tqdm import tqdm

import typing as tp
from imc_analysis.types import Path
from imc_analysis.logging import *

import pandas as pd
import numpy as np
import scanpy as sc
import anndata

import scipy

# import warnings
# warnings.simplefilter("ignore", UserWarning)

def grouped_mwu_test(
    adata: anndata.AnnData,
    condition_keys: list = None,
    multicomp: str = 'BH',
    verbose: bool = False
) -> anndata.AnnData:
    """ Apply MWU test for (cell type density / marker mean expression) matrix either on ROI or patient level

    Args:
        adata (anndata.AnnData): cell type density or marker mean expression matrix on either ROI or patient level.
        condition_keys (list, optional): list of conditions to apply test. Runs all comparisons on obs layer if none provided.
        multicomp (str, optional): Multiple hypothesis correction method. Defaults to BH for Benjamini-Hochberg.

    Returns:
        anndata.AnnData: _description_
    """
    logger.info("Applying Double sided Mann-Whitney U test.")
    import pingouin as pg

    for key in condition_keys:
        assert key in adata.obs.columns
            
    if condition_keys == None:
        condition_keys = adata.obs.columns
        
        # if not categorical, make into categorical variable
        if not isinstance(adata.obs[key].dtype, pd.CategoricalDtype):
            logger.debug(f"Replacing {key} as a categorical variable")
            adata.obs[key] = pd.Categorical(adata.obs[key])
            
            if len(adata.obs[key].cat.categories) > 10:
                logger.warning(f'there is more than 10 categories in key: {key}')
    
    density = adata.to_df()
    adata.uns['mwu'] = dict()
    
    pval_df = pd.DataFrame(columns = ['celltype', 'condition', 'pair', 'U-val', 'p-val'])
    # 'adj. p-val'
    
    for ct in density.columns:
        for cond in condition_keys:

            from itertools import combinations
            comb = list(combinations(adata.obs[cond].cat.categories, 2))

            pd.options.mode.chained_assignment = None  # default='warn'
            #pvals = pd.concat([pg.mwu(density[adata.obs[cond] == p1][ct].tolist(), density[adata.obs[cond] == p2][ct].tolist()) for p1, p2 in comb])
            pval_list = []
            for p1, p2 in comb:
                x = density[adata.obs[cond] == p1][ct].tolist()
                y = density[adata.obs[cond] == p2][ct].tolist()
                
                if len(x) and len(y):
                    pval_list.append(pg.mwu(x, y))
            pvals = pd.concat(pval_list)
            
            mini_pval_df = pvals[['U-val','p-val']]
            mini_pval_df['celltype'] = ct
            mini_pval_df['condition'] = cond
            mini_pval_df['pair'] = comb
            pval_df = pd.concat([pval_df, mini_pval_df])
            
            # BH_pvals = pg.multicomp(pvals['p-val'].tolist(), method = multicomp)
            # BH_pvals = pd.DataFrame({'Significant': BH_pvals[0], 'adj. p-val': BH_pvals[1]}, index = comb)

            # if verbose:
            #     logger.info(BH_pvals)
    def pval_to_star(pval):
        if pval < 0.0001:
            return ' **** '
        elif pval < 0.001:
            return ' *** '
        elif pval < 0.01:
            return ' ** '
        elif pval < 0.05:
            return ' * '
        else:
            return ' ns '
    def pval_to_sci_not(pval):
        return "{:.2E}".format(pval)

    BH_pvals = pg.multicomp(pval_df['p-val'].tolist(), method = multicomp)  
    pval_df['adj. p-val'] = BH_pvals[1]
    pval_df['significance'] = pval_df['adj. p-val'].apply(pval_to_star)
    pval_df['adj. p-val sci'] = pval_df['adj. p-val'].apply(pval_to_sci_not)
    pval_df.reset_index(drop = True, inplace = True)
    adata.uns['mwu'] = pval_df

def celltype_density(
    data: anndata.AnnData,
    celltype: str = 'celltype',
    condition_keys: list = [],
    roi_key: str = 'roi',
    **kwargs: tp.Any,
) -> anndata.AnnData:
    """ Compute cell density for a given celltype in IMC data. Return density matrix as AnnData with specified conditions from the original AnnData.
    
    Args:
        data (anndata.AnnData): Cell Matrix in AnnData with celltyping and roi information in obs layer. Each observation should be a cell. 
        celltype (str, optional): Celltype key to be used. Defaults to 'celltype'.
        condition_keys (list, optional): Condition keys from AnnData obs to store in the return AnnData. Defaults to [].
        roi_key (str, optional): ROI key to be used. Defaults to 'roi'.
        
        y_coord (str, optional): Y coordinate of cells. Defaults to 'Y_centroid'.
        x_coord (str, optional): X coordinate of cells. Defaults to 'X_centroid'.
        
    Returns:
        anndata.AnnData: Cell density matrix as AnnData. Each sample is an ROI, and each column is a cell type.
    """
    logger.info("Calculating cell type density.")
    
    # check for necessary inputs
    assert celltype in data.obs, f"Error: '{celltype}' missing from `data.obs` layer."
    assert roi_key in data.obs, f"Error: '{roi_key}' missing from `data.obs` layer."
    for k in condition_keys:
        assert k in data.obs, f"Error: '{k}' missing from `data.obs` layer."
        
    # Impute ROI area
    if 'ROI_area' not in data.obs:
        impute_roi_area(
            data,
            roi_key = roi_key,
            overwrite = True,
            **kwargs,
        )
    
    keys = [roi_key, 'ROI_area'] + condition_keys
    
    # retrieve cell count matrix
    cell_counts = data.obs.groupby([roi_key,celltype]).count()['sample'].reset_index().pivot(index = roi_key, columns = celltype, values = 'sample')
    # retreive metadata for new obs layer
    data_meta = data.obs[keys].drop_duplicates()
    
    logger.debug(cell_counts)
    logger.debug(data_meta)
    # assert cell count matches metadata length
    assert len(cell_counts) == len(data_meta), 'Error: non-matching cell counts and metadata length. Check for whether condition is unique per sample.'
    
    # set index as roi_key
    data_meta = data_meta.set_index(roi_key)
    
    # create AnnData Object
    density = anndata.AnnData(X = cell_counts.astype(np.float32), obs = data_meta.loc[cell_counts.index])
    density.var.index = cell_counts.columns

    if roi_key in density.obs:
        density.obs = density.obs.set_index(roi_key)

    density.X = density.X / np.array(density.obs['ROI_area'])[:, None]
    
    for cond in condition_keys:
        if f'{cond}_colors' in data.uns:
            density.uns[f'{cond}_colors'] = data.uns[f'{cond}_colors']
    
    return density

def impute_roi_area(
    data: anndata.AnnData,
    roi_key: str = 'roi',
    roi_area_key: str = 'ROI_area',
    y_coord: str = 'Y_centroid',
    x_coord: str = 'X_centroid',
    overwrite: bool = False,
    return_data: bool = False,
    **kwargs: tp.Any,
) -> tp.Optional[anndata.AnnData]:
    """ Function to store imputed ROI area to `obs` layer of AnnData from x and y coordinates.

    Args:
        data (anndata.AnnData): Input AnnData object
        roi_key (str, optional): Identifier from roi. Defaults to 'roi'.
        roi_area_key (str, optional): ROI area key to save the result. Defaults to 'ROI_area'.
        y_coord (str, optional): Y coordinate of cells. Defaults to 'Y_centroid'.
        x_coord (str, optional): X coordinate of cells. Defaults to 'X_centroid'.
        
        overwrite (bool, optional): Whether to overwrite existing roi area. Defaults to False.
        return_data (bool, optional): Whether to return a copy of the data. The original AnnData is modified by default when this is set as False. Defaults to False.

    Returns:
        anndata.AnnData: AnnData object with stored ROI area. Optionally returned when return_data is turned on.
    """
    logger.info(f'Imputing ROI area for data')
    assert roi_key in data.obs, f"Error: {roi_key} not in 'data.obs'"
    
    if return_data:
        data = data.copy()
        
    if not overwrite and roi_area_key in data.obs:
        logger.debug(f'Function not in overwrite mode and {roi_area_key} found in data.obs')
        return

    if overwrite:
        logger.info(f"Overwriting existing {roi_area_key}.")
    
    # Ensure x and y coordinate is in data.obs
    assert y_coord in data.obs, f"Error: {y_coord} not in 'data.obs'"
    assert x_coord in data.obs, f"Error: {x_coord} not in 'data.obs'"
    
    # If spatial missing from obsm, create it based on x and y coord
    if 'spatial' not in data.obsm:
        logger.info(f"Storing `obs[[{y_coord}, {x_coord}]]` to `obsm['spatial']`.")
        data.obsm['spatial'] = np.array(data.obs[[y_coord, x_coord]])

    # cells / mm^2
    data.obs[roi_area_key] = data.obsm['spatial'][:,0] * data.obsm['spatial'][:,1] / 1e6
    # maximum projection per ROI
    data.obs[roi_area_key] = data.obs.groupby(roi_key)[roi_area_key].transform('max')
    
    if return_data:
        return data

def patient_density(
    adata: anndata.AnnData,
    patient_key: str = 'ID',
    celltype_key: str = 'celltype',
    condition_keys: tp.Union[list,None] = None,
    roi_key: str = 'roi'
)->anndata.AnnData:
    """ Calculates celltype density at patient level. 

    Args:
        adata (anndata.AnnData): Cell matrix
        patient_key (str, optional): Patient identifier in obs layer of adata. Defaults to 'ID'.
        celltype_key (str, optional): Cell type identifier in obs layer of adata. Defaults to 'celltype'.
        roi_key (str, optional): Roi identifier in obs layer of adata. Defaults to 'roi'.
        condition_keys (tp.Union[list,None], optional): List of conditions to store in the output anndata obs layer. 
    Returns:
        anndata.AnnData: Cell type density on patient level. Each samples are patient, and each columns are cell types.
    """
    logger.debug(f'celltype key: {celltype_key}')
    logger.debug(f'adata: {adata}')
    
    assert patient_key in adata.obs, f'{patient_key} not in adata.obs layer.'
    if patient_key not in condition_keys:
        condition_keys = condition_keys + [patient_key]
        
    density = celltype_density(
        adata,
        celltype = celltype_key,
        roi_key = roi_key,
        condition_keys = condition_keys
    )
    
    # convert density to count by multiplying area per ROI
    density_df = density.to_df()
    count = density_df * np.array(density.obs['ROI_area'])[:,None]
    count[patient_key] = density.obs[patient_key]
    count = count.groupby(patient_key).sum() # store total roi area
    #count = count.set_index(patient_key)

    features = (density.obs.groupby(patient_key).nunique() <= 1).all()
    patient_metadata = density.obs[features[features].index.tolist() + [patient_key]].drop_duplicates()
    patient_metadata = patient_metadata.set_index(patient_key)

    # create anndata with 
    res = anndata.AnnData(X = count, obs = patient_metadata.loc[count.index])
    res.obs['ROI_area'] = density.obs.groupby(patient_key)['ROI_area'].sum().tolist()
    res.X = res.X / np.array(res.obs['ROI_area'])[:,None]

    # density = adata.to_df()
    # count = density * np.array(adata.obs[roi_area])[:,None]
    # count[patient_key] = adata.obs[patient_key]
    # count = count.groupby(patient_key).sum()
    # #count = count.set_index(patient_key)

    # features = (adata.obs.groupby(patient_key).nunique() <= 1).all()
    # patient_metadata = adata.obs[features[features].index.tolist() + [patient_key]].drop_duplicates()
    # patient_metadata = patient_metadata.set_index(patient_key)

    # res = anndata.AnnData(X = count, obs = patient_metadata.loc[count.index])
    # res.obs[roi_area] = adata.obs.groupby(patient_key)[roi_area].sum().tolist()

    #return res
    return res
  
def grouped_obs_mean(
    adata: anndata.AnnData,
    group_key: str,
    layer=None,
    gene_symbols=None) -> np.array:
    if layer is not None:
        getX = lambda x: x.layers[layer]
    else:
        getX = lambda x: x.X
    if gene_symbols is not None:
        new_idx = adata.var[idx]
    else:
        new_idx = adata.var_names

    grouped = adata.obs.groupby(group_key)
    out = pd.DataFrame(
        np.zeros((adata.shape[1], len(grouped)), dtype=np.float64),
        columns=list(grouped.groups.keys()),
        index=adata.var_names
    )

    for group, idx in grouped.indices.items():
        X = getX(adata[idx])
        out[group] = X.mean(axis=0, dtype=np.float64).tolist()
    return out.T


def compute_mean(
    data: anndata.AnnData,
    condition_keys: list = [],
    obs_key: str = 'roi',
) -> anndata.AnnData:
    """ Compute mean intensity of all markers for a cell matrix grouped by obs_key.

    Args:
        data (anndata.AnnData): cell matrix
        condition_keys (list, optional): list of conditions to keep. Defaults to [].
        obs_key (str, optional): _description_. Defaults to 'roi'.

    Returns:
        anndata.AnnData: Mean expression AnnData. Each sample is on obs_key, and each column is var from input AnnData.
    """
    logger.info(f'Computing mean grouped by: {obs_key}')
    keys = [obs_key] + condition_keys
    for k in keys:
        assert(k in data.obs)

    mean = grouped_obs_mean(data, group_key = obs_key)
    data_meta = data.obs[keys].drop_duplicates()
    logger.debug(mean)
    logger.debug(data_meta)
    # assert cell count matches metadata length
    assert len(mean) == len(data_meta), 'Error: non-matching groups of obs and metadata. Check for whether condition is unique per observation.'
    
    # set index as obs_key
    data_meta = data_meta.set_index(obs_key)
    # mean.index = data_meta.index

    mean_adata = anndata.AnnData(X = mean.astype(np.float32), obs = data_meta.loc[mean.index])
    mean_adata.var.index = mean.columns

    if obs_key in mean_adata.obs:
        mean_adata.obs = mean_adata.obs.set_index(obs_key)
    
    return mean_adata
